//
// greydns.c
//
// GreyDNS - Greylisting DNS-Server
// Copyright (C) 2007  Robert Schulze (rob@rob-schulze.de)
// http://www.rob-schulze.de/publikationen/frickelworx/greydns
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
///////////////////////////////////////////////////////////////////////////////

#include "greydns.h"

int daemonize(void)
{
int pid=fork();
if(pid>0)
	{
	exit(0);
	}
return 1;
}

///////////////////////////////////////////////////////////////////////////////
// sighandler_term
// wird bei SIGTERM aufgerufen und sorgt dafür, dass greydns sauber aussteigt
///////////////////////////////////////////////////////////////////////////////

void sighandler_term()
{
sigset_t mask_set;
signal(SIGTERM,sighandler_term);
signal(SIGINT,sighandler_term);

// andere Signale blocken
//
sigfillset(&mask_set);
sigprocmask(SIG_SETMASK,&mask_set,NULL);

// greydns aus dem mainloop aussteigen lassen
//
RUN=0;
}

int main(void)
{
TXT_GREYLISTED_LEN=(unsigned char) strlen(TXT_GREYLISTED);
printf("started with PID %d\n",getpid());

DBTYPE *db=db_open(DBNAME);
if(NULL==db)
	{
	return(EXIT_SUCCESS); // wie heißen die anderen return values?
	}

if(SIG_ERR==signal(SIGTERM,sighandler_term))
	{
	perror("unable install sighandler for SIGTERM");
	db_close(db);
	return(EXIT_SUCCESS);
	}
if(SIG_ERR==signal(SIGINT,sighandler_term))
	{
	perror("unable install sighandler for SIGINT");
	db_close(db);
	return(EXIT_SUCCESS);
	}

//daemonize();

int sock=sockbind();
fd_set select_fds;
struct timeval select_timeout;

FD_ZERO(&SELECT_FDS);
FD_SET(sock,&SELECT_FDS);

DNS_RR_IN_A=prepack_rr_in_a(&SIZE_DNS_RR_IN_A);
DNS_RR_IN_TXT=prepack_rr_in_txt(&SIZE_DNS_RR_IN_TXT);

while(RUN==1)
	{
	// das Timeout und die FDs muessen jedes Mal neu initialisiert werden!
	//
	memcpy(&select_timeout,&SELECT_INTERVAL,SIZE_TV);
	memcpy(&select_fds,&SELECT_FDS,sizeof(fd_set));

	int select_result=select(sock+1,&select_fds,NULL,NULL,&select_timeout);
	if(select_result==-1)
		{
		if(errno!=EINTR) // nur wenn nicht durch ein Signal unterbrochen
			{
			perror("select failed");
			}
		}
	else if(select_result)
		{
		greydns(sock,db);
		}
	}

free(DNS_RR_IN_A);
free(DNS_RR_IN_TXT);

close(sock);
db_close(db);


printf("exiting gracefully\n");
return(EXIT_SUCCESS);
}

///////////////////////////////////////////////////////////////////////////////
// greydns
// die Hauptfunktion zum Entgegennehmen und Beantworten der Anfrage
///////////////////////////////////////////////////////////////////////////////

void greydns(int sock, DBTYPE *db)
{
ssize_t recvd=0;
struct sockaddr peer;
socklen_t peersize=sizeof(peer);

void *rq=calloc(sizeof(char),512);

recvd=recvfrom(sock,rq,512,0,&peer,&peersize);
if(recvd==-1)
	{
	perror("recvfrom failed");
	free(rq);
	return;
	}

// Antwort zurechtstutzen
//
//printf("greydns: recvd %d bytes\n",recvd);
rq=realloc(rq,recvd);
if(rq==NULL)
	{
	return perror("realloc failed");
	}

int sizeofanswer=recvd;
void *answer=create_answer(rq,&sizeofanswer,db);
if(answer==NULL)
	{
	printf("main: create_answer failed!\n");
	}
else
	{
	int sent=sendto(sock,answer,sizeofanswer,0,&peer,peersize);
	if(sent==-1)
		{	
		perror("send failed");
		}
	free(answer);
	}
}

///////////////////////////////////////////////////////////////////////////////
// create_answer
// erzeugt die Rohdaten der DNS-Antwort
///////////////////////////////////////////////////////////////////////////////

void *create_answer(void *dns, int *sizeofanswer, DBTYPE *db)
{
struct dns_question *question=unpack_request_question(dns);
if(question==NULL)
	{
	return NULL;
	}

int match=db_get(db,question->qname);
int value=time(NULL);
int dosave=1;

void *answer=NULL;
if(match==0)
	{
	answer=pack_response_localhost(dns,question->qname,sizeofanswer);
	value+=GREY_DELAY;
	}
else
	{
	int now=time(NULL);

	if(match > now)
		{
		// weiterer Request vor Ablauf von GREY_DELAY
		//
		answer=pack_response_localhost(dns,question->qname,sizeofanswer);
		dosave=0;
		}
	else if( (match+REMOVE_DELAY) < now)
		{
		// letzte Anfrage zu lange her
		// neu greylisten
		//
		value+=GREY_DELAY;
		answer=pack_response_localhost(dns,question->qname,sizeofanswer);
		}
	else
		{
		// letzte Abfrage im Rahmen
		// greylisting erneuern
		//
		answer=pack_response_nxdomain(dns,sizeofanswer);
		}
	}

if(dosave==1)
	{
	(void)db_set(db,question->qname,value);
	}

free(question->qname);
free(question);

return answer;
}

///////////////////////////////////////////////////////////////////////////////
// pack_response_nxdomain
// erzeugt eine NXDOMAIN-Antwort auf Basis eines dns_packet
///////////////////////////////////////////////////////////////////////////////

void *pack_response_nxdomain(void *dns, int *size)
{
modify_header_nxdomain(dns);
return dns;
}

///////////////////////////////////////////////////////////////////////////////
// modify_header_nxdomain
///////////////////////////////////////////////////////////////////////////////

void modify_header_nxdomain(const void *dns)
{
u_int16_t *ptr=(u_int16_t *)dns;
*(ptr+1)=htons(128 | 3);
}

///////////////////////////////////////////////////////////////////////////////
// pack_response_localhost
// erzeugt eine IN A 127.0.0.2 und TXT Antwort auf Basis eines dns_packet
///////////////////////////////////////////////////////////////////////////////

void *pack_response_localhost(void *dns, char *domainname, int *size)
{
modify_header_localhost(dns);

int domainname_size=0;
void *packed_dn=pack_domainname(domainname,&domainname_size);

unsigned short appendix_size=domainname_size+SIZE_DNS_RR_IN_A+domainname_size+SIZE_DNS_RR_IN_TXT;

dns=realloc(dns,*size+appendix_size);

if(dns!=NULL)
	{
	memset(dns+*size,0,appendix_size);

	// IN A
	dns=memcpy(dns+*size,packed_dn,domainname_size);
	dns=memcpy(dns+domainname_size,DNS_RR_IN_A,SIZE_DNS_RR_IN_A);

	// IN TXT
	dns=memcpy(dns+SIZE_DNS_RR_IN_A,packed_dn,domainname_size);
	dns=memcpy(dns+domainname_size,DNS_RR_IN_TXT,SIZE_DNS_RR_IN_TXT);

	*size=*size+appendix_size;
	dns-=*size-SIZE_DNS_RR_IN_TXT; // Zeiger zuruecksetzen
	}
else
	{
	perror("realloc failed");
	}

free(packed_dn);
if(dns==NULL)
	{
	return NULL;
	}
return(dns);
}



///////////////////////////////////////////////////////////////////////////////
// modify_header_localhost
///////////////////////////////////////////////////////////////////////////////

void modify_header_localhost(const void *dns)
{
u_int16_t *ptr=(u_int16_t *)dns;
*(ptr+1)=htons(128);
*(ptr+3)=htons(2);
}


///////////////////////////////////////////////////////////////////////////////
// prepack_rr_in_txt
// erzeugt den statischen Teil des IN TXT Resource Records
// zum globalen Festhalten
///////////////////////////////////////////////////////////////////////////////

void *prepack_rr_in_txt(unsigned short *size)
{
void *rr=NULL;

u_int16_t type=htons(16);
u_int16_t class=htons(1);
u_int32_t ttl=htonl(0);
u_int16_t rdlength=htons(TXT_GREYLISTED_LEN+SIZE_8);

*size=((3*SIZE_16)+SIZE_32+SIZE_8+TXT_GREYLISTED_LEN);

rr=malloc(*size);
memset(rr,0,*size);

rr=memcpy(rr,&type,SIZE_16);
rr=memcpy(rr+SIZE_16,&class,SIZE_16);
rr=memcpy(rr+SIZE_16,&ttl,SIZE_32);
rr=memcpy(rr+SIZE_32,&rdlength,SIZE_16);
rr=memcpy(rr+SIZE_16,&TXT_GREYLISTED_LEN,SIZE_8);
rr=memcpy(rr+SIZE_8,TXT_GREYLISTED,TXT_GREYLISTED_LEN);


rr-=*size-TXT_GREYLISTED_LEN; // Zeiger wieder auf den Anfang setzen

return rr;
}

///////////////////////////////////////////////////////////////////////////////
// prepack_rr_in_a
// erzeugt den statischen Teil des IN A Resource Records
// zum globalen Festhalten
///////////////////////////////////////////////////////////////////////////////

void *prepack_rr_in_a(unsigned short *size)
{
void *rr=NULL;

struct in_addr addr;
if(0==inet_aton(A_GREYLISTED,&addr))
	{
	perror("inet_aton failed");
	return NULL;
	}

u_int16_t type=htons(1);
u_int16_t class=htons(1);
u_int32_t ttl=htonl(0);
u_int16_t rdlength=sizeof(struct in_addr);

*size=(3*SIZE_16)+SIZE_32+sizeof(struct in_addr);

rr=malloc(*size);
memset(rr,0,*size);

rr=memcpy(rr,&type,SIZE_16);
rr=memcpy(rr+SIZE_16,&class,SIZE_16);
rr=memcpy(rr+SIZE_16,&ttl,SIZE_32);
rr=memcpy(rr+SIZE_32,&rdlength,SIZE_16);
rr=memcpy(rr+SIZE_16,&addr,sizeof(struct in_addr));

rr-=*size-sizeof(struct in_addr); // Zeiger wieder auf den Anfang setzen

return rr;
}


///////////////////////////////////////////////////////////////////////////////
// pack_domainname
// packt einen Domainnamen aus der dns-Struktur 
///////////////////////////////////////////////////////////////////////////////

void *pack_domainname(const char *domainname, int *size)
{
void *buf=NULL;
char *ptr=(char *)domainname;

unsigned char lsize=0;
int bufsize=0;
char done=0;

while(done==0)
	{	
	char *dotat=index(ptr,'.');
	if(dotat!=NULL)
		{
		lsize=(dotat-ptr);
		//printf("found dot at offset %d\n",lsize);
		
		buf=realloc(buf,bufsize+lsize+SIZE_8);
		memcpy(buf+bufsize,&lsize,SIZE_8);
		memcpy(buf+bufsize+SIZE_8,ptr,lsize);

		bufsize+=lsize+SIZE_8;
		ptr=dotat+1;
		lsize=0;
		}
	else
		{
		done=1;
		}
	}

// letztes Label anhaengen
//
buf=realloc(buf,bufsize+SIZE_8);
lsize=0;
memcpy(buf+bufsize,&lsize,SIZE_8);
bufsize+=1;

*size+=bufsize;
return buf;
}

///////////////////////////////////////////////////////////////////////////////
// unpack_request_question
// packt den gefragten Domainnamen und Type,Class aus
///////////////////////////////////////////////////////////////////////////////

struct dns_question *unpack_request_question(const void *data)
{
char *bindata=(char *)data;
bindata+=SIZE_HDR;

int labelsize=0;
int strsize=0;
char *qname=NULL;

do
	{
	unsigned char size=0;
	memcpy(&size,bindata,SIZE_8);
	
	bindata+=SIZE_8;
	if(size>0)
		{
		qname=realloc(qname,strsize+size+1+1); // qname + label + '.' + 0
		memset(qname+strsize,'\0',size+1+1);
		memcpy(qname+strsize,bindata,size);
		
		*(qname+strsize+size)='.';
		strsize+=size+1;
		}

	labelsize=size;
	bindata+=size;
	}
while(labelsize!=0);
if(qname==NULL)
	{
	return NULL;
	}

//printf("qname start: %p end %p (%p)\n",qname,qname+strsize,qname+strlen(qname));

struct dns_question *q;
q=malloc(sizeof(struct dns_question));
if(q==NULL)
	{
	perror("malloc failed");
	free(qname);
	return(NULL);
	}


if( 	(NULL==memcpy(&q->qtype,bindata,SIZE_16)) || 
	(NULL==memcpy(&q->qclass,bindata+SIZE_16,SIZE_16)) )
	{
	perror("memcpy failed");
	free(q);
	free(qname);
	return NULL;
	}

q->qtype=ntohs(q->qtype);
q->qclass=ntohs(q->qclass);
q->qname=qname;

//printf("unpack_request_question:\n\tqname %s\n\tqclass %d\n\tqtype %d\n",q->qname,q->qclass,q->qtype);

return q;
}

///////////////////////////////////////////////////////////////////////////////
// int sockbind
// baut das Socket, ruft bind auf und gibt das Socket zurueck
///////////////////////////////////////////////////////////////////////////////

int sockbind(void)
{
struct in_addr addr;

inet_aton(LISTEN,&addr);

int sock=socket(PF_INET,SOCK_DGRAM,0);
if(sock==-1)
	{
	perror("unable to create socket");
	return 0;
	}

struct sockaddr_in serv_addr;
serv_addr.sin_family = AF_INET;
serv_addr.sin_port = htons(PORT);
serv_addr.sin_addr = addr;

if(bind(sock,(struct sockaddr *) &serv_addr,sizeof(serv_addr))==-1)
	{
	perror("unable to bind socket");
	return 0;
	}
return sock;
}



